



    <div id="page" class="innerpage clearfix">
    <h1>Search Results</h1>
    {% if search.performed %}

     {% paginate search.results by 10 %}

     {% if search.results == empty %}
      <div class="latest-news">Your search for "{{search.terms | escape}}" did not yield any results</div>
        {% else %}


    <ul class="search-list clearfix">
     {% for item in search.results %}
      <li>
          <h3 class="stitle">{{ item.title | link_to: item.url }}</h3>
          <p class="sinfo">{{ item.content | strip_html | truncatewords: 65 | highlight: search.terms }} ... <a href="{{item.url}}" title="">view this item</a></p>
      </li>
    {% endfor %}
    </ul>
     {% endif %}

    <div class="paginate clearfix">
      {{ paginate | default_pagination }}
    </div>


    <div id="three-reasons" class="clearfix">
      <h3>Why Shop With Us?</h3>
      <ul>
        <li class="two-a">
          <h4>24 Hours</h4>
          <p>We're always here to help.</p>
        </li>
        <li class="two-c">
          <h4>No Spam</h4>
          <p>We'll never share your info.</p>
        </li>
        <li class="two-d">
          <h4>Secure Servers</h4>
          <p>Checkout is 256bit encrypted.</p>
        </li>
      </ul>
    </div>
  </div>

{% endpaginate %}
{% endif %}
