  <div id="page" class="innerpage clearfix">
    <div id="text-page">
      <h1>Post from our blog...</h1>
      {% paginate blog.articles by 20 %}
          {% for article in blog.articles  %}

          <div class="entry">
            <h1><span><a href="{{ article.url }}">{{ article.title }}</a></span></h1>
            <div class="entry-post">
              <div class="meta">{{ article.created_at | date: "%b %d" }}</div>
              {{ article.content }}
            </div>
          </div>

        {% endfor %}

        <div class="paginate clearfix">
            {{ paginate | default_pagination }}
        </div>

      {% endpaginate %}
    </div>

    <div id="three-reasons" class="clearfix">
      <h3>Why Shop With Us?</h3>
      <ul>
        <li class="two-a">
          <h4>24 Hours</h4>
          <p>We're always here to help.</p>
        </li>
        <li class="two-c">
          <h4>No Spam</h4>
          <p>We'll never share your info.</p>
        </li>
        <li class="two-d">
          <h4>Secure Servers</h4>
          <p>Checkout is 256bit encrypted.</p>
        </li>
      </ul>
    </div>
  </div>
