<div id="home-page">
  <h3 class="heading-shaded">Featured products...</h3>
  <div class="featured-prod-row clearfix">
    {% for product in collections.frontpage.products %}
      <div class="featured-prod-item">
          <p>
        <a href="{{product.url}}"><img src="{{ product.featured_image | product_img_url: 'small' }}" alt="{{ product.title | escape }}"/></a>
        </p>
        <h4><a href="{{product.url}}">{{product.title}}</a></h4>
        {% if product.compare_at_price %}
          {% if product.price_min != product.compare_at_price %}
            <p class="prd-price">Was:<del>{{product.compare_at_price | money}}</del></p>
            <p class="prd-price"><ins>Now: {{product.price_min | money}}</ins></p>
          {% endif %}
        {% else %}
          <p class="prd-price"><ins>{{product.price_min | money}}</ins></p>
        {% endif %}
      </div>
    {% endfor %}
  </div>

  <div id="articles">
    {% assign article = pages.frontpage %}
    {% if article.content != "" %}
      <h3>{{ article.title }}</h3>
      {{ article.content }}
    {% else %}
      In <em>Admin &gt; Blogs &amp; Pages</em>, create a page with the handle <strong><code>frontpage</code></strong> and it will show up here.<br />
      {{ "Learn more about handles" | link_to: "http://wiki.shopify.com/Handle" }}
    {% endif %}
  </div>
</div>
